/*
Calibreで生成したEPUB2ファイルをEPUB3ファイルに変換するプログラム
2022.01.24 Takeshi Yoneki
ostra@asahi.email.ne.jp

確認環境
Rust 1.62.0 (Windows 10 21H2)
Rust 1.62.0 (Ubuntu 20.04)

Cargo.toml
[dependencies]
zip = "0.6"

epub2to3_RSから独自簡易HTMLパーサを利用するように変更。

2022.11.15 Rust 1.65.0
*/

mod html_parser;
mod char_entity;

use std::env;
use std::fs;
use std::io::prelude::*;
use crate::html_parser::Node;
use zip::read::ZipArchive;	
use zip::write::ZipWriter;
use std::io::BufReader;
use std::io::BufWriter;
use std::path::Path;

#[derive(Debug)]
struct Config {
    vertical: bool,
    src: String,
    dst: String,
    css: String,
    css_text: String,
    id: bool,
}

impl Config {
    fn new() -> Config {
        Config {
            vertical: false, src: String::new(), dst: String::new(), css: String::new(),
            css_text: String::new(), id: true,
        }
    }

    fn is_true(s: &String) -> bool {
        let t = &s.to_lowercase();
        t == "true" || t == "ok" || t == "on" || t == "t" || t == "go" || t == "good" || t == "yes"
    }

    fn parse_param(&mut self, args: &[String]) {
        if args.len() < 2 {
            return;
        }
        let p = &args[0];
        let v = &args[1];
        match &p[..] {
            "--direction" | "-r"  => self.vertical = v.to_lowercase() == "v",
            "--src" | "-s" | "--input" | "-i" => self.src = v.clone(),
            "--dst" | "-d" | "--output" | "-o" => self.dst = v.clone(),
            "--css" | "-c" => self.css = String::from(v),
            "--id" | "--kobo" | "-k" => self.id = Config::is_true(v),
            _ => {
                self.parse_param(&args[1..]);
                return;
            },
        }
        self.parse_param(&args[2..]);
        return;
    }

    fn load_css(&mut self) {
        if self.css.len() > 0 {
            let mut f = fs::File::open(&self.css).unwrap();
            f.read_to_string(&mut self.css_text).unwrap();
        }
    }
}

fn has_svg_html(cfg: &Config, text: &String) -> bool {
    let doc = html_parser::parse(text);
    let svg_elems = doc.get_elements_by_tag_name("svg");
    svg_elems.len() > 0
}

fn get_svg_html_path_list(cfg: &Config, in_path: &String) -> Vec<String> {
    let mut inz = ZipArchive::new(BufReader::new(fs::File::open(in_path).unwrap())).unwrap();
    let mut names: Vec<String> = Vec::new();
    for index in 0..inz.len() {
        let mut entry = inz.by_index(index).unwrap();
        if entry.is_file() {
            let name = String::from(entry.name());
            let n = name.to_lowercase();
            if n.ends_with(".html") || n.ends_with(".xhtml") {
                let mut text: String = String::new();
                entry.read_to_string(&mut text).unwrap();
                if has_svg_html(cfg, &text) {
                    names.push(name);
                }
            }
        }
    }
    names
}

fn get_ncx_path(elem: &Node) -> String {
    let href = elem.get_elements_by_tag_name("manifest").iter().fold(
        String::new(), 
        |acc, manifest_elem| {
            if acc.is_empty() {
                manifest_elem.get_elements_by_tag_name("item").iter().fold(
                    String::new(), 
                    |acc, item_elem| {
                        if acc.is_empty() {
                            if item_elem.get_attribute("media-type") == String::from("application/x-dtbncx+xml") {
                                return item_elem.get_attribute("href");
                            }
                        }
                        return acc;
                    }
                )
            }
            else {
                acc
            }
        }
    );
    href
}

fn get_nav_path_from_ncx_path(path: &String) -> String {
    let path = Path::new(path);
    String::from(path.parent().unwrap().join(Path::new("nav.html")).to_str().unwrap())
}

fn fix_opf(cfg: &Config, text: &String, svg_html_list: &Vec<String>) -> String {
    let doc = html_parser::parse(text);
    //let doc = cfg.parser.parse(text);
    let root_elem = doc.get_document_element();
    let ncx_path = get_ncx_path(&root_elem);
    let d = doc.rule_transformer(
        &mut |node: Node, _| -> Node {
            if node.is_element() {
                let tag = node.get_tag_name();
                let tag = tag.split(":").last().unwrap();
                match &tag[..] {
                    "package" => {
                        return node.set_attribute("version", "3.0");
                    },
                    "creator" | "contributor" => {
                        return node.remove_attribute_all();
                    }
                    "identifier" => {
                        if let Node::Element(tag, mut attr, child) = node {
                            let ax: Vec<(String, String)> = attr.drain(..).filter(
                                |(n, _)| {
                                    !(n.split(":").last().unwrap() == "scheme")
                                }
                            ).collect();
                            return Node::Element(tag, ax, child);
                        }
                    },
                    "date" => {
                        let d: String = String::from(&node.get_text_content()[0..19]) + "Z";
                        return Node::Element(String::from("meta"), vec![(String::from("property"), String::from("dcterms:modified"))], vec![Node::PCData(d)]);
                    },
                    "manifest" => {
                        if ncx_path.len() > 0 {
                            let e = Node::Element(
                                String::from("item"),
                                vec![
                                    (String::from("href"), get_nav_path_from_ncx_path(&ncx_path)),
                                    (String::from("id"), String::from("nav")),
                                    (String::from("media-type"), String::from("application/xhtml+xml")),
                                    (String::from("properties"), String::from("nav")),
                                ],
                                Vec::new()
                            );
                            return node.append_child(e);
                        }
                    },
                    "spine" => {
                        if cfg.vertical {
                            return node.set_attribute("page-progression-direction", "rtl");
                        }
                    },
                    "item" => {
                        let h = node.get_attribute("href");
                        if svg_html_list.contains(&h) {
                            return node.set_attribute("properties", "svg");
                        }
                    },
                    _ => {
                    },
                }
            }
            return node;
        }
    );
    //
    let d2 = d.pretty("\t");
    d2.to_xml_string()
}

fn conv_ncx_to_nav(cfg: &Config, text: &String) -> String {
    //
    fn get_grandchild_text_value(elem: &Node, tag1: &str, tag2: &str) -> String {
        elem.get_elements_by_tag_name(tag1).iter().fold(
            String::new(), 
            |acc, elem1| {
                if acc.is_empty() {
                    elem1.get_elements_by_tag_name(tag2).iter().fold(
                        String::new(), 
                        |acc, elem2| {
                            if acc.is_empty() {
                                elem2.get_text_content()
                            }
                            else {
                                acc
                            }
                        }
                    )
                }
                else {
                    acc
                }
            }
        )
    }
    //
    fn get_child_attr_value(elem: &Node, tag: &str, attr_name: &str) -> String {
        elem.get_elements_by_tag_name(tag).iter().fold(
            String::new(),
            |acc, elem1| {
                if acc.is_empty() {
                    elem1.get_attribute(attr_name)
                }
                else {
                    acc
                }
            }
        )
    }
    //
    let doc = html_parser::parse(text);
    let root_elem = doc.get_document_element();
    let title = get_grandchild_text_value(root_elem, "docTitle", "text");
    let mut np_list: Vec<(String, String)> = {
        root_elem.get_elements_by_tag_name("navPoint").iter().filter_map(
            |n| {
                let np_title = get_grandchild_text_value(n, "navLabel", "text");
                let np_src = get_child_attr_value(n, "content", "src");
                if np_title.is_empty() || np_src.is_empty() {
                    None
                }
                else {
                    Some((np_title, np_src))
                }
            }
        ).collect()
    };
    let tmp_text = "<html xmlns=\"http://www.w3.org/1999/xhtml\" xmlns:epub=\"http://www.idpf.org/2007/ops\"><head><meta charset=\"UTF-8\" /><title /></head><body><nav epub:type=\"toc\"><h1 /><ol /></nav></body></html>";
    let tmp_doc = html_parser::parse(tmp_text);
    let d = tmp_doc.rule_transformer(
        &mut |node: Node, _| -> Node {
            if node.is_element() {
                let tag = node.get_tag_name();
                match &tag[..] {
                    "title" | "h1" => {
                        if let Node::Element(tag, attr, _) = node {
                            return Node::Element(tag, attr, vec![Node::PCData(String::from(&title))]);
                        }
                    },
                    "ol" => {
                        let mut li_elems: Vec<Node> = np_list.drain(..).map(
                            |(np_title, np_src)| {
                                let a_elem = Node::Element(
                                    String::from("a"), 
                                    vec![(String::from("href"), np_src)],
                                    vec![Node::PCData(np_title)]
                                );
                                let li_elem = Node::Element(
                                    String::from("li"), 
                                    vec![],
                                    vec![a_elem]
                                );
                                li_elem
                            }
                        ).collect();
                        if let Node::Element(tag, attr, mut child) = node {
                            child.append(&mut li_elems);
                            return Node::Element(tag, attr, child);
                        }
                    },
                    _ => {
                    },
                }
            }
            return node;
        }
    );
    let d2 = d.pretty("\t");
    d2.to_xml_string()
}

fn fix_css(cfg: &Config, text: &String) -> String {
  let t1 = "\n\n@page { margin: 10%; }\n";
  let t2 = "html { -epub-writing-mode: vertical-rl; writing-mode: vertical-rl; }\n";
  let t3 = &cfg.css_text;
  if cfg.vertical {
      String::from(text) + t1 + t2 + t3
  }
  else {
      String::from(text) + t1 + t3
  }
}

fn fix_html(cfg: &Config, text: &String) -> String {
    let doc = html_parser::parse(text);
    let mut id_number = 1;
    let d = doc.rule_transformer(
        &mut |node: Node, _| -> Node {
            if node.is_element() {
                let tag = node.get_tag_name();
                match &tag[..] {
                    "meta" => {
                        if !node.get_attribute("http-equiv").is_empty() {
                            return Node::Empty;
                        }
                    },
                    "link" => {
                        if node.get_attribute("rel") == String::from("alternate stylesheet") {
                            return Node::Empty;
                        }
                    },
                    "p" | "div" | "span" | "h1" | "h2" | "a" | "img" => {
                        if cfg.id {
                            if node.get_attribute("id").is_empty() {
                                let id = String::from("kobo.") + &id_number.to_string() + ".1";
                                id_number += 1;
                                return node.set_attribute("id", &id)
                            }
                        }
                    },
                    _ => {
                    }
                }
            }
            return node;
        }
    );
    let d2 = d.pretty("\t");
    d2.to_xml_string()
}

fn fix_zip(cfg: &Config, in_path: &String, out_path: &String) {
    let svg_html_list = get_svg_html_path_list(cfg, in_path);
    let mut inz = ZipArchive::new(BufReader::new(fs::File::open(in_path).unwrap())).unwrap();
    let mut outz = ZipWriter::new(BufWriter::new(fs::File::create(out_path).unwrap()));
    for index in 0..inz.len() {
        let mut entry = inz.by_index(index).unwrap();
        let n = entry.name().to_lowercase();
        if n.ends_with(".opf") {
            outz.start_file(String::from(entry.name()), zip::write::FileOptions::default()).unwrap();
            let mut text: String = String::new();
            entry.read_to_string(&mut text).unwrap();
            outz.write(fix_opf(cfg, &text, &svg_html_list).as_bytes()).unwrap();
        }
        else if n.ends_with(".ncx") {
            outz.start_file(get_nav_path_from_ncx_path(&String::from(entry.name())), zip::write::FileOptions::default()).unwrap();
            let mut text: String = String::new();
            entry.read_to_string(&mut text).unwrap();
            outz.write(conv_ncx_to_nav(cfg, &text).as_bytes()).unwrap();
            //
            outz.start_file(String::from(entry.name()), zip::write::FileOptions::default()).unwrap();
            outz.write(text.as_bytes()).unwrap();
        }
        else if n.ends_with(".css") {
            outz.start_file(String::from(entry.name()), zip::write::FileOptions::default()).unwrap();
            let mut text: String = String::new();
            entry.read_to_string(&mut text).unwrap();
            outz.write(fix_css(cfg, &text).as_bytes()).unwrap();
        }
        else if n.ends_with(".html") || n.ends_with(".xhtml") {
            outz.start_file(String::from(entry.name()), zip::write::FileOptions::default()).unwrap();
            let mut text: String = String::new();
            entry.read_to_string(&mut text).unwrap();
            outz.write(fix_html(cfg, &text).as_bytes()).unwrap();
        }
        else if Path::new(&n).file_name().unwrap().to_str().unwrap() == "mimetype" {
            outz.start_file(String::from(entry.name()), zip::write::FileOptions::default().compression_method(zip::CompressionMethod::Stored)).unwrap();
            let mut text: String = String::new();
            entry.read_to_string(&mut text).unwrap();
            outz.write(text.as_bytes()).unwrap();
        }
        else {
            outz.raw_copy_file(entry).unwrap();
        }
    }
}

fn main() {
    let mut cfg = Config::new();
    let args: Vec<String> = env::args().collect();
    cfg.parse_param(&args[1..]);
    cfg.load_css();
    fix_zip(&cfg, &cfg.src, &cfg.dst);
}

